﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.IO.Ports;
using System.IO;
using System.Threading;

namespace espfilemanager_cs
{
    public partial class FormMain : Form
    {
        int prange, psize, pfree;
        bool intransit;

        public FormMain()
        {
            InitializeComponent();
            switchControls(false);
            listserialports();
        }

        delegate void switchControlsCallback(bool enb, bool onlyfilectl);
        private void switchControls(bool enb, bool onlyfilectl = false)
        {
            if (buttonDel.InvokeRequired)
            {
                Invoke(new switchControlsCallback(switchControls), new object[] { enb, onlyfilectl });
                return;
            }

            if (enb)
            {
                buttonDel.Enabled = true;
                buttonUp.Enabled = true;
                buttonDown.Enabled = true;
                listBoxFile.Enabled = true;
                if (!onlyfilectl)
                {
                    comboBoxCom.Enabled = false;
                    buttonComupdate.Enabled = false;
                }
            }
            else
            {
                buttonDel.Enabled = false;
                buttonUp.Enabled = false;
                buttonDown.Enabled = false;
                listBoxFile.Enabled = false;
                if (!onlyfilectl)
                {
                    comboBoxCom.Enabled = true;
                    buttonComupdate.Enabled = true;
                }
            }
        }

        private void listserialports()
        {
            string[] ports = SerialPort.GetPortNames();
            for (int i = 0; i < ports.Length; i++)
            {
                comboBoxCom.Items.Add(ports[i]);
            }
        }

        delegate void AppendTextCallback(string txt, bool ret);
        private void AppendText(String txt, bool ret = true)
        {
            if (textBoxSerial.InvokeRequired)
            {
                Invoke(new AppendTextCallback(AppendText), new object[] { txt, ret });
                //AppendTextCallback d = new AppendTextCallback(AppendText);
                //BeginInvoke(d, new object[] { txt, ret });
                return;
            }
            textBoxSerial.AppendText(txt);

            if (ret)
                textBoxSerial.AppendText("\r\n");
        }

        private void buttonComupdate_Click(object sender, EventArgs e)
        {
            comboBoxCom.Items.Clear();
            listserialports();
        }

        private void buttonConnect_Click(object sender, EventArgs e)
        {
            //https://learn.microsoft.com/ja-jp/visualstudio/ide/create-csharp-winform-visual-studio?view=vs-2022
            //https://learn.microsoft.com/ja-jp/dotnet/api/system.io.ports.serialport?view=dotnet-plat-ext-7.0
            //https://qiita.com/mag2/items/d15bc3c9d66ce0c8f6b1
            //http://www.ys-labo.com/CS%20Tips/2008/CS%20Tips%20RS232C.html
            //https://www.atsumitakeshi.com/csharp_arduino1.html
            //https://learn.microsoft.com/en-us/dotnet/api/system.io.ports.serialport.datareceived?view=dotnet-plat-ext-7.0

            if (intransit) return;

            if (serialPortEsp.IsOpen)
            {
                sendSerial("mode 0");
                serialPortEsp.Close();
                listBoxFile.Items.Clear();
                AppendText("Port closed.");
                buttonConnect.Text = "接続";
                switchControls(false);
                setprogresspos(0);
                return;
            }

            if (comboBoxCom.SelectedIndex < 0) return;

            serialPortEsp.PortName = comboBoxCom.SelectedItem.ToString(); //"COM8";
            serialPortEsp.BaudRate = 115200;
            serialPortEsp.DataBits = 8;
            serialPortEsp.Parity = Parity.None;
            serialPortEsp.StopBits = StopBits.One;
            serialPortEsp.Handshake = Handshake.None;
            //serialPortEsp.RtsEnable = true; //required for arduino but not for esp
            //serialPortEsp.DtrEnable = true;
            serialPortEsp.NewLine = "\n";

            try
            {
                serialPortEsp.Open();
                if (serialPortEsp.IsOpen)
                {
                    AppendText("Serial port OK");
                    buttonConnect.Text = "切断";
                    sendSerial("mode 5");
                    AppendText("Change mode...");
                    switchControls(true);
                }
                else
                    AppendText("Failed to open port");
            }
            catch (Exception ex)
            {
                AppendText("Failed to open port");
            }
        }

        private void buttonClose_Click(object sender, EventArgs e)
        {
            this.Close();
        }

        private void buttonDel_Click(object sender, EventArgs e)
        {
            if (intransit) return;
            if (listBoxFile.SelectedIndex >= 0)
            {
                if (MessageBox.Show(listBoxFile.SelectedItem.ToString() + " を削除してよろしいですか？", "ファイルの削除", MessageBoxButtons.OKCancel, MessageBoxIcon.Warning) == DialogResult.OK)
                {
                    sendSerial("del " + (listBoxFile.SelectedIndex + 1));
                }
            }
        }

        private void buttonDown_Click(object sender, EventArgs e)
        {
            if (intransit) return;
            if (listBoxFile.SelectedIndex >= 0)
            {
                sendSerial("down " + (listBoxFile.SelectedIndex + 1));
            }
        }

        private void buttonUp_Click(object sender, EventArgs e)
        {
            if (intransit) return;
            if (openFileDialog1.ShowDialog() == DialogResult.OK)
            {
                sendSerialFile(openFileDialog1.FileName);
            }
        }

        private void listBoxFile_DoubleClick(object sender, EventArgs e)
        {
            if (intransit) return;
            if (listBoxFile.SelectedIndex >= 0)
            {
                sendSerial("down " + (listBoxFile.SelectedIndex + 1));
            }
        }

        private void FormMain_DragEnter(object sender, DragEventArgs e)
        {
            if (e.Data.GetDataPresent(DataFormats.FileDrop))

            {
                e.Effect = DragDropEffects.Copy;

            }
        }

        private void FormMain_DragDrop(object sender, DragEventArgs e)
        {
            if (intransit) return;
            string[] files = (string[])e.Data.GetData(DataFormats.FileDrop);
            sendSerialFile(files[0]);
        }

        delegate void SetProgressCallback(int newpos);
        private void setprogresspos(int newpos)
        {
            if (progressBarTrans.InvokeRequired)
            {
                Invoke(new SetProgressCallback(setprogresspos), new object[] { newpos });
            }
            else
            {
                progressBarTrans.Value = newpos;
            }
        }

        delegate void SetProgressmaxCallback(int newpos);
        private void setprogressmax(int newpos)
        {
            if (progressBarTrans.InvokeRequired)
            {
                Invoke(new SetProgressmaxCallback(setprogressmax), new object[] { newpos });
            }
            else
            {
                progressBarTrans.Maximum = newpos;
            }
        }


        private void sendSerial(string msg)
        {
            if (serialPortEsp.IsOpen)
            {
                //serialPortEsp.WriteLine(msg);
                byte[] b_sjis = Encoding.GetEncoding("shift-jis").GetBytes(msg+"\n");
                serialPortEsp.Write(b_sjis, 0, b_sjis.Length);
            }
        }

        delegate void saveSerialFileCallback();
        private void saveSerialFile()
        {
            if (textBoxSerial.InvokeRequired)
            {
                Invoke(new saveSerialFileCallback(saveSerialFile));
                return;
            }

            AppendText("File receive complete");
            progressBarTrans.Maximum = prange;
            progressBarTrans.Value = psize;
            saveFileDialog1.FileName = serialfilename;
            if (saveFileDialog1.ShowDialog() == DialogResult.OK)
            {
                //https://araramistudio.jimdo.com/2018/12/20/c-%E3%81%A7%E3%83%90%E3%82%A4%E3%83%8A%E3%83%AA%E3%83%95%E3%82%A1%E3%82%A4%E3%83%AB%E3%81%AB%E6%9B%B8%E3%81%8D%E8%BE%BC%E3%82%80/
                try
                {
                    var fileout = new BinaryWriter(new FileStream(saveFileDialog1.FileName, FileMode.Create));
                    fileout.Write(serialfiledata, 0, serialfiledatasize);
                    fileout.Close();
                }
                catch (Exception e)
                {
                    AppendText("File create error");
                    MessageBox.Show("ファイルの保存に失敗しました", "エラー", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
            }
        }

        private void sendSerialFile(string fname)
        {
            if (intransit) return;
            if (!serialPortEsp.IsOpen)
            {
                AppendText("Serial not connected");
                return;
            }
            if (!File.Exists(fname))
            {
                MessageBox.Show("ファイルが見つからないかアクセスできません", "エラー", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }
            intransit = true;
            switchControls(false, true);

            //https://koshinran.hateblo.jp/entry/2018/03/05/171356
            FileInfo fi = new FileInfo(fname);
            string filename = Path.GetFileName(fname);
            AppendText("Upoading file " + filename + "\r\nsize:" + fi.Length);

            if (fi.Length / 1024 >= pfree)
            {
                MessageBox.Show("空き容量が足りません 空き：" + pfree + "kb サイズ：" + (fi.Length / 1024) + "kb", "エラー", MessageBoxButtons.OK, MessageBoxIcon.Error);
                intransit = false;
                return;
            }

            sendSerial("up\n" + filename + "\n" + fi.Length);
            progressBarTrans.Maximum = (int)fi.Length;
            progressBarTrans.Value = 0;

            Thread.Sleep(2000); //need this for esp to reply "ok"

            int sendsize;
            byte[] fileBytes = File.ReadAllBytes(fname);

            int sentbyte = 0;
            try
            {
                while (sentbyte < fileBytes.Length)
                {
                    sendsize = Math.Min(fileBytes.Length - sentbyte, 512);
                    serialPortEsp.Write(fileBytes, sentbyte, sendsize);
                    sentbyte += sendsize;
                    progressBarTrans.Value = sentbyte;

                }
            }
            catch (Exception e)
            {
                AppendText("Serial write error: " + e.Message);
            }

            intransit = false;
            switchControls(true, true);
            progressBarTrans.Maximum = prange;
            progressBarTrans.Value = psize;
        }

        enum res
        {
            none,
            mode,
            list,
            err,
            file,
            filesize,
            filedata,
            msg
        }
        byte[] serialbuff_b = new byte[32768];
        int serialbuff_b_idx = 0;
        //string serialbuff;
        string serialfilename;
        byte[] serialfiledata;
        int serialfiledatasize, serialfiledatapos;
        bool newresponse;
        res responsemode;
        int last_serialreceive;

        private void serialPortEsp_DataReceived(object sender, SerialDataReceivedEventArgs e)
        {
            //throw old data
            int current_time;
            current_time = System.Environment.TickCount;
            if (current_time - last_serialreceive > 5000)
            {
                responsemode = res.none;
                //serialbuff = "";
                serialbuff_b_idx = 0;
                Array.Clear(serialbuff_b,0,serialbuff_b.Length);
            }
            last_serialreceive = current_time;

            //if file data, get in binary mode
            if (responsemode == res.filedata)
            {

                while (serialPortEsp.BytesToRead > 0)
                {
                    serialfiledata[serialfiledatapos++] = (byte)serialPortEsp.ReadByte();
                    if (serialfiledatapos >= serialfiledatasize)
                    {
                        responsemode = res.none;
                        //serialbuff = "";
                        serialbuff_b_idx = 0;
                        Array.Clear(serialbuff_b, 0, serialbuff_b.Length);
                        saveSerialFile();
                        switchControls(true, true);
                        return;
                    }
                    setprogresspos(serialfiledatapos);
                }
                return;
            }

            int bytestoread = serialPortEsp.BytesToRead;
            byte[] ser_buff = new byte[bytestoread];
            serialPortEsp.Read(ser_buff, 0, bytestoread);
            analyzeSerialResponse(ser_buff);

            /*string data = serialPortEsp.ReadExisting();   // ポートから文字列を受信する
            if (!string.IsNullOrEmpty(data))
                serialReceive(data);*/
            /*string data = serialPortEsp.ReadLine();
            serialReceive(data);*/
        }

        delegate void analyzeSerialResponseCallback(byte[] msg);
        private void analyzeSerialResponse(byte[] msg)
        {
            if (textBoxSerial.InvokeRequired)
            {
                Invoke(new analyzeSerialResponseCallback(analyzeSerialResponse), new object[] { msg });
                return;
            }

            for (int i = 0; i < msg.Length; i++)
            {
                if (responsemode != res.filedata && msg[i] == '\n')
                {
                    string serialbuff;
                    serialbuff_b[serialbuff_b_idx] = 0;
                    //https://www.ipentec.com/document/csharp-bytearray-to-string
                    //serialbuff = Encoding.ASCII.GetString(serialbuff_b);
                    //serialbuff = Encoding.GetEncoding("shift_jis").GetString(serialbuff_b);
                    //serialbuff = Encoding.UTF8.GetString(serialbuff_b);
                    //serialbuff = serialbuff.TrimEnd('\0');
                    //serialbuff = Encoding.UTF8.GetString(serialbuff_b).TrimEnd('\0');
                    serialbuff = Encoding.GetEncoding("shift_jis").GetString(serialbuff_b).TrimEnd('\0');

                    if (responsemode != res.file && responsemode != res.filesize && responsemode != res.filedata)
                    {
                        if (serialbuff.Length > 1)
                            AppendText(serialbuff);
                    }

                    if (serialbuff == "-----")
                    {
                        newresponse = true;
                        serialbuff = "";
                        serialbuff_b_idx = 0;
                        Array.Clear(serialbuff_b, 0, serialbuff_b.Length);
                        responsemode = res.none;
                    }
                    else
                    {
                        if (newresponse)
                        {
                            if (serialbuff == "mode")
                                responsemode = res.mode;
                            else if (serialbuff == "list")
                            {
                                responsemode = res.list;
                                listBoxFile.Items.Clear();
                            }
                            else if (serialbuff == "error")
                                responsemode = res.err;
                            else if (serialbuff == "file")
                                responsemode = res.file;
                            else if (serialbuff == "message")
                                responsemode = res.msg;
                            //else
                            //AppendText("Unknown response: " + serialbuff);
                        }
                        else
                        {
                            /* response example
                                -----
                                list
                                1:/test.txt  SIZE: 10
                                1 file(s) / 1345kb total / 0kb used / 1345kb free
                             */

                            if (responsemode == res.mode)
                            {
                                if (serialbuff.Contains("5"))
                                {
                                    AppendText("Switch mode ok");
                                    sendSerial("list");
                                }
                                responsemode = res.none;
                            }
                            else if (responsemode == res.list)
                            {
                                if (serialbuff.Contains(":"))
                                {
                                    int fsize = int.Parse(serialbuff.Substring(serialbuff.LastIndexOf(" ")));
                                    string fname = serialbuff.Substring(serialbuff.IndexOf("/") + 1);
                                    fname = fname.Substring(0, fname.LastIndexOf("  SIZE"));
                                    listBoxFile.Items.Add(fname + " [" + fsize + "]");
                                }
                                else if (serialbuff.Contains("kb"))
                                {
                                    string strsub;
                                    strsub = serialbuff.Substring(0, serialbuff.IndexOf("kb total"));
                                    prange = int.Parse(strsub.Substring(strsub.LastIndexOf(" ") + 1));
                                    strsub = serialbuff.Substring(0, serialbuff.IndexOf("kb used"));
                                    psize = int.Parse(strsub.Substring(strsub.LastIndexOf(" ") + 1));
                                    pfree = prange - psize;
                                    progressBarTrans.Maximum = prange;
                                    progressBarTrans.Value = psize;
                                    responsemode = res.none;
                                }
                            }
                            else if (responsemode == res.msg)
                            {
                                responsemode = res.none;
                            }
                            else if (responsemode == res.err)
                            {
                                responsemode = res.none;
                            }
                            else if (responsemode == res.file)
                            {
                                serialfilename = serialbuff;
                                AppendText("Name: " + serialfilename);
                                responsemode = res.filesize;
                            }
                            else if (responsemode == res.filesize)
                            {
                                serialfiledatasize = int.Parse(serialbuff);
                                AppendText("Size: " + serialfiledatasize);
                                if (serialfiledatasize > 0)
                                {
                                    Array.Resize<byte>(ref serialfiledata, serialfiledatasize);
                                    serialfiledatapos = 0;
                                    responsemode = res.filedata;
                                    progressBarTrans.Maximum = serialfiledatasize;
                                    progressBarTrans.Value = serialfiledatapos;
                                    i++;
                                    switchControls(false, true);
                                }
                                else
                                {
                                    responsemode = res.none;
                                }
                            }
                        }
                        serialbuff = "";
                        serialbuff_b_idx = 0;
                        Array.Clear(serialbuff_b, 0, serialbuff_b.Length);
                        newresponse = false;
                    }
                }

                if (responsemode != res.filedata)
                {
                    if (msg[i] != '\r' && msg[i] != '\n' && serialbuff_b_idx < 32768)
                        serialbuff_b[serialbuff_b_idx++] = msg[i];
                }
                else
                {
                    serialfiledata[serialfiledatapos++] = (byte)msg[i];
                    progressBarTrans.Value = serialfiledatapos;
                    if (serialfiledatapos >= serialfiledatasize)
                    {
                        saveSerialFile();
                        switchControls(true, true);
                    }
                }
            }
        }
    }
}
